# Copyright (C) 2020-2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pathlib import Path
from unittest.mock import MagicMock

import tomlkit

from pontos.version import VersionCommand


class UpdatePyprojectVersionTestCase(unittest.TestCase):
    def test_empty_pyproject_toml(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.read_text.return_value = ""

        cmd = VersionCommand(pyproject_toml_path=fake_path)

        cmd.update_pyproject_version('20.04dev1')

        text = fake_path.write_text.call_args[0][0]

        toml = tomlkit.parse(text)

        self.assertEqual(toml['tool']['poetry']['version'], '20.4.dev1')

    def test_empty_tool_section(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.read_text.return_value = "[tool]"

        cmd = VersionCommand(pyproject_toml_path=fake_path)
        cmd.update_pyproject_version('20.04dev1')

        text = fake_path.write_text.call_args[0][0]

        toml = tomlkit.parse(text)

        self.assertEqual(toml['tool']['poetry']['version'], '20.4.dev1')

    def test_empty_tool_poetry_section(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.read_text.return_value = "[tool.poetry]"

        cmd = VersionCommand(pyproject_toml_path=fake_path)
        cmd.update_pyproject_version('20.04dev1')

        text = fake_path.write_text.call_args[0][0]

        toml = tomlkit.parse(text)

        self.assertEqual(toml['tool']['poetry']['version'], '20.4.dev1')

    def test_override_existing_version(self):
        fake_path_class = MagicMock(spec=Path)
        fake_path = fake_path_class.return_value
        fake_path.read_text.return_value = '[tool.poetry]\nversion = "1.2.3"'

        cmd = VersionCommand(pyproject_toml_path=fake_path)
        cmd.update_pyproject_version('20.04dev1')

        text = fake_path.write_text.call_args[0][0]

        toml = tomlkit.parse(text)

        self.assertEqual(toml['tool']['poetry']['version'], '20.4.dev1')



# Copyright (C) 2020 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from pathlib import Path
from .__version__ import __version__

from .version import (
    PontosVersionCommand,
    VersionCommand,
    VersionError,
    safe_version,
    strip_version,
    is_version_pep440_compliant,
    get_version_from_pyproject_toml,
)
from .cmake_version import CMakeVersionParser, CMakeVersionCommand
from .helper import (
    calculate_calendar_version,
    get_current_version,
    get_next_patch_version,
    get_next_dev_version,
)


def main(leave=True, args=None):
    available_cmds = [
        ('CMakeLists.txt', CMakeVersionCommand),
        ('pyproject.toml', PontosVersionCommand),
    ]
    for file_name, cmd in available_cmds:
        project_definition_path = Path.cwd() / file_name
        if project_definition_path.exists():
            result = cmd().run(args)
            if leave:
                sys.exit(result)
            return result == 0, file_name
    if leave:
        sys.exit("No command found")
    return False, ""


__all__ = [
    '__version__',
    'calculate_calendar_version',
    'CMakeVersionCommand',
    'get_version_from_pyproject_toml',
    'get_current_version',
    'get_next_patch_version',
    'get_next_dev_version',
    'is_version_pep440_compliant',
    'PontosVersionCommand',
    'safe_version',
    'strip_version',
    'VersionCommand',
    'VersionError',
]



# Copyright (C) 2020-2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import struct
import re

from io import StringIO
from subprocess import CompletedProcess, CalledProcessError

from pathlib import Path
from unittest import TestCase
from unittest.mock import patch
from argparse import Namespace

from pontos.updateheader.updateheader import (
    _get_modified_year as get_modified_year,
    _find_copyright as find_copyright,
    _add_header as add_header,
    _update_file as update_file,
    _parse_args as parse_args,
    main,
)


class UpdateHeaderTestCase(TestCase):
    def setUp(self):
        self.args = Namespace()
        self.args.company = 'Greenbone Networks GmbH'

        self.path = Path(__file__).parent

        self.regex = re.compile(
            "[Cc]opyright.*?(19[0-9]{2}|20[0-9]{2}) "
            f"?-? ?(19[0-9]{{2}}|20[0-9]{{2}})? ({self.args.company})"
        )

    @patch('pontos.updateheader.updateheader.run')
    def test_get_modified_year(self, run_mock):
        test_file = self.path / "test.py"
        test_file.touch(exist_ok=True)

        run_mock.return_value = CompletedProcess(
            args=[
                'git',
                'log',
                '-1',
                '--format=%ad',
                '--date=format:%Y',
                f'{test_file}',
            ],
            returncode=0,
            stdout='2020\n',
            stderr='',
        )

        year = get_modified_year(f=test_file)
        self.assertEqual(year, '2020')

        test_file.unlink()

    def test_get_modified_year_error(self):
        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        with self.assertRaises(CalledProcessError):
            get_modified_year(f=test_file)

    def test_find_copyright(self):
        test_line = "# Copyright (C) 1995-2021 Greenbone Networks GmbH"
        test2_line = "# Copyright (C) 1995 Greenbone Networks GmbH"
        invalid_line = (
            "# This program is free software: "
            "you can redistribute it and/or modify"
        )

        # Full match
        found, match = find_copyright(regex=self.regex, line=test_line)
        self.assertTrue(found)
        self.assertIsNotNone(match)
        self.assertEqual(match['creation_year'], "1995")
        self.assertEqual(match['modification_year'], "2021")
        self.assertEqual(match['company'], self.args.company)

        # No modification Date
        found, match = find_copyright(regex=self.regex, line=test2_line)
        self.assertTrue(found)
        self.assertIsNotNone(match)
        self.assertEqual(match['creation_year'], "1995")
        self.assertEqual(match['modification_year'], None)
        self.assertEqual(match['company'], self.args.company)

        # No match
        found, match = find_copyright(regex=self.regex, line=invalid_line)
        self.assertFalse(found)
        self.assertIsNone(match)

    def test_add_header(self):
        expected_header = """# -*- coding: utf-8 -*-
# Copyright (C) 2020 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

        header = add_header(
            suffix=".py", licence='AGPL-3.0-or-later', company=self.args.company
        )

        self.assertEqual(header, expected_header)

    def test_add_header_wrong_file_suffix(self):

        with self.assertRaises(ValueError):
            add_header(
                suffix=".prr",
                licence='AGPL-3.0-or-later',
                company=self.args.company,
            )

    def test_add_header_licence_not_found(self):

        with self.assertRaises(FileNotFoundError):
            add_header(
                suffix=".py",
                licence='AAAGPL-3.0-or-later',
                company=self.args.company,
            )

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_file_not_existing(self, mock_stdout):
        self.args.year = '2020'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        with self.assertRaises(FileNotFoundError):
            update_file(file=test_file, regex=self.regex, args=self.args)

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: File is not existing.\n",
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_file_wrong_licence(self, mock_stdout):
        self.args.year = '2020'
        self.args.changed = False
        self.args.licence = 'AAAGPL-3.0-or-later'

        test_file = self.path / "test.py"
        test_file.touch()

        code = update_file(file=test_file, regex=self.regex, args=self.args)
        self.assertEqual(code, 1)

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: Licence file for "
            "AAAGPL-3.0-or-later is not existing.\n",
        )

        test_file.unlink()

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_file_suffix_invalid(self, mock_stdout):
        self.args.year = '2020'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        test_file = self.path / "test.pppy"
        test_file.touch()

        code = update_file(file=test_file, regex=self.regex, args=self.args)
        self.assertEqual(code, 1)

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: No licence header for the format .pppy found.\n",
        )

        test_file.unlink()

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_file_binary_file(self, mock_stdout):
        self.args.year = '2020'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        # create a Binary file ...
        # https://stackoverflow.com/a/30148554
        with open(test_file, 'wb') as f:
            f.write(struct.pack('>if', 42, 2.71828182846))

        with self.assertRaises(UnicodeDecodeError):
            code = update_file(file=test_file, regex=self.regex, args=self.args)
            self.assertEqual(code, 1)

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: Ignoring binary file.\n",
        )

        test_file.unlink()

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_file_changed(self, mock_stdout):
        self.args.year = '1995'
        self.args.changed = True
        self.args.licence = 'AGPL-3.0-or-later'

        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        with self.assertRaises(FileNotFoundError):
            code = update_file(file=test_file, regex=self.regex, args=self.args)
            self.assertEqual(code, 1)

        ret = mock_stdout.getvalue()

        self.assertEqual(
            ret,
            f"{test_file}: Could not get date of last "
            f"modification using git, using {self.args.year} instead.\n"
            f"{test_file}: File is not existing.\n",
        )

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_create_header(self, mock_stdout):
        self.args.year = '1995'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        expected_header = """# -*- coding: utf-8 -*-
# Copyright (C) 2020 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

        test_file = self.path / "test.py"
        test_file.touch()

        # test_file.write_text(expected_header)

        code = update_file(file=test_file, regex=self.regex, args=self.args)

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: Added licence header.\n",
        )
        self.assertEqual(code, 0)
        self.assertEqual(expected_header, test_file.read_text())
        test_file.unlink()

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_header_in_file(self, mock_stdout):
        self.args.year = '2021'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        header = """# -*- coding: utf-8 -*-
# Copyright (C) 2020 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        test_file.write_text(header)

        code = update_file(file=test_file, regex=self.regex, args=self.args)

        self.assertEqual(code, 0)
        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: Changed Licence Header "
            "Copyright Year None -> 2021\n",
        )
        self.assertIn(
            '# Copyright (C) 2020-2021 Greenbone Networks GmbH',
            test_file.read_text(),
        )

        test_file.unlink()

    @patch('sys.stdout', new_callable=StringIO)
    def test_update_header_ok_in_file(self, mock_stdout):
        self.args.year = '2021'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        header = """# -*- coding: utf-8 -*-
# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

        test_file = self.path / "test.py"
        if test_file.exists():
            test_file.unlink()

        test_file.write_text(header)

        code = update_file(file=test_file, regex=self.regex, args=self.args)

        self.assertEqual(code, 0)
        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            f"{test_file}: Licence Header is ok.\n",
        )
        self.assertIn(
            '# Copyright (C) 2021 Greenbone Networks GmbH',
            test_file.read_text(),
        )

        test_file.unlink()

    def test_argparser_files(self):
        self.args.year = '2021'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        args = ['-f', 'test.py', '-y', '2021', '-l', 'AGPL-3.0-or-later']

        args = parse_args(args)
        self.assertIsNotNone(args)
        self.assertEqual(args.company, self.args.company)
        self.assertEqual(args.files, ['test.py'])
        self.assertEqual(args.year, self.args.year)
        self.assertEqual(args.licence, self.args.licence)

    def test_argparser_dir(self):
        self.args.year = '2020'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'

        args = ['-d', '.', '-c', '-l', 'AGPL-3.0-or-later']

        args = parse_args(args)
        self.assertIsNotNone(args)
        self.assertEqual(args.company, self.args.company)
        self.assertEqual(args.directory, '.')
        self.assertTrue(args.changed)
        self.assertEqual(args.year, '2021')
        self.assertEqual(args.licence, self.args.licence)

    @patch('pontos.updateheader.updateheader._parse_args')
    def test_main(self, argparser_mock):
        self.args.year = '2021'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'
        self.args.files = ['test.py']
        self.args.directory = None

        argparser_mock.return_value = self.args

        code = main()

        # I have no idea how or why test main ...
        self.assertEqual(code, 0)

    @patch('sys.stdout', new_callable=StringIO)
    @patch('pontos.updateheader.updateheader._parse_args')
    def test_main_never_happen(self, argparser_mock, mock_stdout):
        self.args.year = '2021'
        self.args.changed = False
        self.args.licence = 'AGPL-3.0-or-later'
        self.args.files = None
        self.args.directory = None

        argparser_mock.return_value = self.args

        # I have no idea how or why test main ...
        with self.assertRaises(SystemExit):
            main()

        ret = mock_stdout.getvalue()
        self.assertEqual(
            ret,
            "Specify files to update!\n",
        )



# -*- coding: utf-8 -*-
# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import unittest
from unittest.mock import patch
from pathlib import Path


from pontos.version import (
    calculate_calendar_version,
    get_next_patch_version,
    get_next_dev_version,
)


class CalculateNextVersionTestCase(unittest.TestCase):
    def test_calculate_calendar_versions(self):
        today = datetime.datetime.today()

        filenames = ['pyproject.toml', 'CMakeLists.txt']
        mocks = [
            'pontos.version.helper.PontosVersionCommand',
            'pontos.version.helper.CMakeVersionCommand',
        ]
        current_versions = [
            '20.4.1.dev3',
            f'{str(today.year % 100)}.4.1.dev3',
            f'19.{str(today.month)}.1.dev3',
            f'{str(today.year % 100)}.{str(today.month)}.1.dev3',
            f'{str(today.year % 100)}.{str(today.month)}.1',
        ]
        assert_versions = [
            f'{str(today.year % 100)}.{str(today.month)}.0',
            f'{str(today.year % 100)}.{str(today.month)}.0',
            f'{str(today.year % 100)}.{str(today.month)}.0',
            f'{str(today.year % 100)}.{str(today.month)}.1',
            f'{str(today.year % 100)}.{str(today.month)}.2',
        ]

        tmp_path = Path.cwd() / 'tmp'

        for filename, mock in zip(filenames, mocks):
            for current_version, assert_version in zip(
                current_versions, assert_versions
            ):
                tmp_path.mkdir(parents=True, exist_ok=True)
                os.chdir(tmp_path)
                proj_file = Path.cwd() / filename
                proj_file.touch()
                with patch(mock) as cmd_mock:
                    cmd_mock.return_value.get_current_version.return_value = (
                        current_version
                    )

                    release_version = calculate_calendar_version()

                    self.assertEqual(release_version, assert_version)

                os.chdir('..')
                proj_file.unlink()

        tmp_path.rmdir()

    def test_get_next_dev_version(self):
        current_versions = [
            '20.4.1',
            '20.4.1',
            '19.1.2',
            '1.1.1',
            '20.6.1',
        ]
        assert_versions = [
            '20.4.2',
            '20.4.2',
            '19.1.3',
            '1.1.2',
            '20.6.2',
        ]

        for current_version, assert_version in zip(
            current_versions, assert_versions
        ):
            next_version = get_next_dev_version(current_version)

            self.assertEqual(assert_version, next_version)

    def test_get_next_patch_version(self):
        today = datetime.datetime.today()

        filenames = ['pyproject.toml', 'CMakeLists.txt']
        mocks = [
            'pontos.version.helper.PontosVersionCommand',
            'pontos.version.helper.CMakeVersionCommand',
        ]
        current_versions = [
            '20.4.1.dev3',
            f'{str(today.year % 100)}.4.1.dev3',
            f'19.{str(today.month)}.1.dev3',
            f'{str(today.year % 100)}.{str(today.month)}.1',
            '20.6.1',
        ]
        assert_versions = [
            '20.4.1',
            f'{str(today.year % 100)}.4.1',
            f'19.{str(today.month)}.1',
            f'{str(today.year % 100)}.{str(today.month)}.2',
            '20.6.2',
        ]

        tmp_path = Path.cwd() / 'tmp'

        for filename, mock in zip(filenames, mocks):
            for current_version, assert_version in zip(
                current_versions, assert_versions
            ):
                tmp_path.mkdir(parents=True, exist_ok=True)
                os.chdir(tmp_path)
                proj_file = Path.cwd() / filename
                proj_file.touch()
                with patch(mock) as cmd_mock:
                    cmd_mock.return_value.get_current_version.return_value = (
                        current_version
                    )

                    release_version = get_next_patch_version()

                    self.assertEqual(release_version, assert_version)

                os.chdir('..')
                proj_file.unlink()

        tmp_path.rmdir()



# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .terminal import Terminal

__term = None  # pylint: disable=invalid-name

__all__ = [
    'error',
    'fail',
    'info',
    'bold_info',
    'ok',
    'out',
    'warning',
]


def ok(message: str) -> None:
    __term.ok(message)


def fail(message: str) -> None:
    __term.fail(message)


def error(message: str) -> None:
    __term.error(message)


def warning(message: str) -> None:
    __term.warning(message)


def info(message: str) -> None:
    __term.info(message)


def bold_info(message: str) -> None:
    __term.bold_info(message)


def out(message: str):
    __term.print(message)


def _set_terminal(term: Terminal):
    global __term  # pylint: disable=global-statement, invalid-name
    __term = term



# Copyright (C) 2020-2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pathlib import Path
from unittest.mock import MagicMock

from pontos.version.version import PontosVersionCommand, VersionError

from . import use_cwd


class PontosVersionCommandTestCase(unittest.TestCase):
    def test_missing_pyproject_toml_file(self):
        with use_cwd(Path(__file__).parent), self.assertRaisesRegex(
            VersionError, r'^.* not found\.$'
        ):
            PontosVersionCommand()

    def test_missing_tool_pontos_version_section(self):
        pyproject_toml_path = MagicMock(spec=Path).return_value
        pyproject_toml_path.exists.return_value = True
        pyproject_toml_path.read_text.return_value = '[tool.pontos]'

        with self.assertRaisesRegex(
            VersionError, r'^\[tool\.pontos\.version\] section missing in .*\.$'
        ):
            PontosVersionCommand(pyproject_toml_path=pyproject_toml_path)

    def test_missing_version_module_file_key(self):
        pyproject_toml_path = MagicMock(spec=Path).return_value
        pyproject_toml_path.exists.return_value = True
        pyproject_toml_path.read_text.return_value = (
            '[tool.pontos.version]\nname="foo"'
        )

        with self.assertRaisesRegex(
            VersionError,
            r'^version-module-file key not set in \[tool\.pontos\.version\] '
            r'section .*\.$',
        ):
            PontosVersionCommand(pyproject_toml_path=pyproject_toml_path)

    def test_with_all_settings(self):
        pyproject_toml_path = MagicMock(spec=Path).return_value
        pyproject_toml_path.exists.return_value = True
        pyproject_toml_path.read_text.return_value = (
            '[tool.pontos.version]\nversion-module-file="foo/__version__.py"'
        )

        cmd = PontosVersionCommand(pyproject_toml_path=pyproject_toml_path)

        self.assertEqual(cmd.version_file_path, Path('foo') / '__version__.py')



# -*- coding: utf-8 -*-
# Copyright (C) 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import unittest
import tempfile
import os
import sys
from pathlib import Path
import shutil


from pontos.release.helper import (
    get_project_name,
    find_signing_key,
    update_version,
)
from pontos import version
from pontos.terminal import (
    _set_terminal,
)
from pontos.terminal.terminal import Terminal


class TestHelperFunctions(unittest.TestCase):
    def setUp(self):
        self.shell_cmd_runner = lambda x: subprocess.run(
            x,
            shell=True,
            check=True,
            errors="utf-8",  # use utf-8 encoding for error output
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _set_terminal(Terminal())

        self.tmpdir = Path(tempfile.gettempdir()) / 'testrepo'
        self.tmpdir.mkdir(parents=True, exist_ok=True)
        self.shell_cmd_runner(f'git -C {self.tmpdir} init')
        self.shell_cmd_runner(
            f'git -C {self.tmpdir} remote add foo https://foo.bar/bla.git'
        )
        self.shell_cmd_runner(
            f'git -C {self.tmpdir} remote add '
            'origin https://foo.bar/testrepo.git'
        )

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir)

    def test_get_project_name(self):
        proj_path = Path.cwd()
        os.chdir(self.tmpdir)
        project = get_project_name(
            shell_cmd_runner=self.shell_cmd_runner, remote='foo'
        )
        self.assertEqual(project, 'bla')

        project = get_project_name(shell_cmd_runner=self.shell_cmd_runner)
        self.assertEqual(project, 'testrepo')
        os.chdir(proj_path)

    def test_find_signing_key(self):
        # save possibly set git signing key from user temporarly
        try:
            saved_key = self.shell_cmd_runner(
                'git config user.signingkey'
            ).stdout.strip()
        except subprocess.CalledProcessError:
            saved_key = None

        self.shell_cmd_runner(
            'git config user.signingkey '
            '1234567890ABCEDEF1234567890ABCEDEF123456'
        )

        signing_key = find_signing_key(shell_cmd_runner=self.shell_cmd_runner)
        self.assertEqual(
            signing_key, '1234567890ABCEDEF1234567890ABCEDEF123456'
        )

        # reset the previously saved signing key ...
        if saved_key is not None:
            self.shell_cmd_runner(f'git config user.signingkey {saved_key}')

    def test_find_no_signing_key(self):
        # save possibly set git signing key from user temporarly
        try:
            saved_key = self.shell_cmd_runner(
                'git config user.signingkey'
            ).stdout.strip()
        except subprocess.CalledProcessError:
            saved_key = None

        try:
            self.shell_cmd_runner('git config --unset user.signingkey')
        except subprocess.CalledProcessError as e:
            self.assertEqual(e.returncode, 5)

        signing_key = find_signing_key(shell_cmd_runner=self.shell_cmd_runner)
        self.assertEqual(signing_key, '')

        # reset the previously saved signing key ...
        if saved_key is not None:
            self.shell_cmd_runner(f'git config user.signingkey {saved_key}')

    def test_update_version_not_found(self):
        proj_path = Path.cwd()
        os.chdir(self.tmpdir)
        executed, filename = update_version(
            to='21.4.4', _version=version, develop=True
        )
        self.assertFalse(executed)
        self.assertEqual(filename, '')

        os.chdir(proj_path)

    def test_update_version_no_version_file(self):
        proj_path = Path.cwd()
        sys.path.append(self.tmpdir)
        os.chdir(self.tmpdir)

        module_path = self.tmpdir / "testrepo"
        module_path.mkdir(parents=True, exist_ok=True)
        init_file = module_path / "__init__.py"
        init_file.touch()
        version_file = module_path / "__version__.py"
        toml = self.tmpdir / "pyproject.toml"
        toml.write_text(
            '[tool.poetry]\n'
            'name = "testrepo"\n'
            'version = "21.6.2.dev1"\n\n'
            '[tool.pontos.version]\n'
            'version-module-file = "testrepo/__version__.py"\n'
        )
        executed, filename = update_version(
            to='21.4.4', _version=version, develop=False
        )
        toml_text = toml.read_text()
        print(toml_text)
        self.assertEqual(filename, 'pyproject.toml')
        self.assertTrue(executed)
        self.assertEqual(
            toml_text,
            '[tool.poetry]\n'
            'name = "testrepo"\n'
            'version = "21.4.4"\n\n'
            '[tool.pontos.version]\n'
            'version-module-file = "testrepo/__version__.py"\n',
        )
        self.assertTrue(version_file.exists())
        version_text = version_file.read_text()
        self.assertEqual(
            version_text,
            '# pylint: disable=invalid-name\n\n'
            '# THIS IS AN AUTOGENERATED FILE. DO NOT TOUCH!\n\n'
            '__version__ = "21.4.4"\n',
        )

        toml.unlink()
        shutil.rmtree(module_path)
        sys.path.remove(self.tmpdir)
        os.chdir(proj_path)



# -*- coding: utf-8 -*-
# Copyright (C) 2020 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import shutil
import unittest

from dataclasses import dataclass
from pathlib import Path
from unittest.mock import MagicMock, patch

import requests

from pontos.release.helper import version
from pontos import release, changelog


@dataclass
class StdOutput:
    stdout: bytes


_shutil_mock = MagicMock(spec=shutil)


@patch("pontos.release.helper.shutil", new=_shutil_mock)
class PrepareTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ['GITHUB_TOKEN'] = 'foo'
        os.environ['GITHUB_USER'] = 'bar'

    def test_prepare_successfully(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

    def test_prepare_calendar_successfully(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--calendar',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

    def test_use_git_signing_key_on_prepare(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--git-signing-key',
            '0815',
            '--release-version',
            '0.0.1',
        ]
        called = []

        def runner(cmd):
            called.append(cmd)
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)
        self.assertIn(
            "git commit -S 0815 -m 'automatic release to 0.0.1'", called
        )
        self.assertIn(
            "git tag -u 0815 v0.0.1 -m 'automatic release to 0.0.1'", called
        )

    def test_fail_if_tag_is_already_taken(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_changelog = MagicMock(spec=changelog)

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('v0.0.1'.encode())
        with self.assertRaises(
            ValueError, msg='git tag v0.0.1 is already taken'
        ):
            release.main(
                shell_cmd_runner=runner,
                _path=fake_path_class,
                _version=fake_version,
                _changelog=fake_changelog,
                leave=False,
                args=args,
            )

    def test_not_release_when_no_project_found(self):
        fake_path_class = MagicMock(spec=Path)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (False, '')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_not_release_when_updating_version_fails(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (False, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')

        args = [
            'prepare',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)


@patch("pontos.release.helper.shutil", new=_shutil_mock)
class ReleaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ['GITHUB_TOKEN'] = 'foo'
        os.environ['GITHUB_USER'] = 'bar'
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    def test_release_successfully(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'release',
            '--release-version',
            '0.0.1',
            '--next-version',
            '0.0.2dev',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

    def test_not_release_successfully_when_github_create_release_fails(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 401
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'release',
            '--release-version',
            '0.0.1',
        ]
        runner = lambda x: StdOutput('')
        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_release_to_specific_git_remote(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'release',
            '--project',
            'foo',
            '--release-version',
            '0.0.1',
            '--next-version',
            '0.0.2.dev1',
            '--git-remote-name',
            'upstream',
        ]

        called = []

        def runner(cmd: str):
            called.append(cmd)
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)

        self.assertIn('git push --follow-tags upstream', called)
        self.assertIn('git add MyProject.conf', called)
        self.assertIn(
            "git commit -m '* Update to version"
            " 0.0.2.dev1\n* Add empty changelog after 0.0.1'",
            called,
        )
        print(called)


@patch("pontos.release.helper.shutil", new=_shutil_mock)
class SignTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ['GITHUB_TOKEN'] = 'foo'
        os.environ['GITHUB_USER'] = 'bar'
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    def test_fail_sign_on_invalid_get_response(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 404
        fake_get.text = self.valid_gh_release_response
        fake_requests.get.return_value = fake_get
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'foo',
            '--release-version',
            '0.0.1',
        ]

        def runner(_: str):
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_fail_sign_on_upload_fail(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 500
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_requests.get.return_value = fake_get
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'foo',
            '--release-version',
            '0.0.1',
        ]

        def runner(_: str):
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertFalse(released)

    def test_successfully_sign(self):
        fake_path_class = MagicMock(spec=Path)
        fake_requests = MagicMock(spec=requests)
        fake_get = MagicMock(spec=requests.Response).return_value
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_requests.get.return_value = fake_get
        fake_post = MagicMock(spec=requests.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        fake_requests.post.return_value = fake_post
        fake_version = MagicMock(spec=version)
        fake_version.main.return_value = (True, 'MyProject.conf')
        fake_changelog = MagicMock(spec=changelog)
        fake_changelog.update.return_value = ('updated', 'changelog')
        args = [
            'sign',
            '--project',
            'bar',
            '--release-version',
            '0.0.1',
        ]

        called = []

        def runner(cmd: str):
            called.append(cmd)
            return StdOutput('')

        released = release.main(
            shell_cmd_runner=runner,
            _path=fake_path_class,
            _requests=fake_requests,
            _version=fake_version,
            _changelog=fake_changelog,
            leave=False,
            args=args,
        )
        self.assertTrue(released)



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020 - 2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import Callable, Dict, List, Tuple, Union
import json
import tempfile
from pathlib import Path
import shutil
import subprocess

import requests

from pontos import version
from pontos.terminal import error, warning, info, ok, out


def build_release_dict(
    release_version: str,
    release_changelog: str,
    *,
    name: str = '',
    target_commitish: str = '',
    draft: bool = False,
    prerelease: bool = False,
) -> Dict[str, Union[str, bool]]:
    """
    builds the dict for release post on github, see:
    https://docs.github.com/en/rest/reference/repos#create-a-release
    for more details.

    Arguments:
        release_version: The version (str) that will be set
        release_changelog: content of the Changelog (str) for the release
        name: name (str) of the release, e.g. 'pontos 1.0.0'
        target_commitish: needed when tag is not there yet (str)
        draft: If the release is a draft (bool)
        prerelease: If the release is a pre release (bool)

    Returns:
        The dictionary containing the release information.
    """
    tag_name = (
        release_version
        if release_version.startswith('v')
        else "v" + release_version
    )
    return {
        'tag_name': tag_name,
        'target_commitish': target_commitish,
        'name': name,
        'body': release_changelog,
        'draft': draft,
        'prerelease': prerelease,
    }


def commit_files(
    filename: str,
    commit_msg: str,
    shell_cmd_runner: Callable,
    *,
    git_signing_key: str = '',
):
    """Add files to staged and commit staged files.

    filename: The filename of file to add and commit
    commit_msg: The commit message for the commit
    shell_cmd_runner: The runner for shell commands
    git_signing_key: The signing key to sign this commit

    Arguments:
        to: The version (str) that will be set
        develop: Wether to set version to develop or not (bool)

    Returns:
       executed: True if successfully executed, False else
       filename: The filename of the project definition
    """

    shell_cmd_runner(f"git add {filename}")
    shell_cmd_runner("git add *__version__.py || echo 'ignoring __version__'")
    shell_cmd_runner("git add CHANGELOG.md")
    if git_signing_key:
        shell_cmd_runner(
            f"git commit -S {git_signing_key} -m '{commit_msg}'",
        )
    else:
        shell_cmd_runner(f"git commit -m '{commit_msg}'")


def download(
    url: str,
    filename: str,
    requests_module: requests,
    path: Path,
) -> Path:
    """Download file in url to filename

    Arguments:
        url: The url of the file we want to download
        filename: The name of the file to store the download in
        requests_module: the python request module
        path: the python pathlib.Path module

    Returns:
       Path to the downloaded file
    """

    file_path = path(tempfile.gettempdir()) / filename

    with requests_module.get(url, stream=True) as resp, file_path.open(
        mode='wb'
    ) as download_file:
        shutil.copyfileobj(resp.raw, download_file)

    return file_path


def get_project_name(
    shell_cmd_runner: Callable,
    *,
    remote: str = 'origin',
) -> str:
    """Get the git repository name

    Arguments:
        shell_cmd_runner: The runner for shell commands
        remote: the remote to look up the name (str) default: origin

    Returns:
        project name
    """
    ret = shell_cmd_runner(f'git remote get-url {remote}')
    return ret.stdout.split('/')[-1].replace('.git', '').strip()


def find_signing_key(shell_cmd_runner: Callable) -> str:
    """Find the signing key in the config

    Arguments:
        shell_cmd_runner: The runner for shell commands

    Returns:
        git signing key or empty string
    """

    try:
        proc = shell_cmd_runner('git config user.signingkey')
    except subprocess.CalledProcessError as e:
        # The command `git config user.signingkey` returns
        # return code 1 if no key is set.
        # So we will return empty string ...
        if e.returncode == 1:
            warning("No signing key found.")
        return ''
    # stdout should return "\n" if no key is available
    # and so git_signing_key should
    # return '' if no key is available ...
    return proc.stdout.strip()


def update_version(
    to: str, _version: version, *, develop: bool = False
) -> Tuple[bool, str]:
    """Use pontos-version to update the version.

    Arguments:
        to: The version (str) that will be set
        _version: Version module
        develop: Wether to set version to develop or not (bool)

    Returns:
       executed: True if successfully executed, False else
       filename: The filename of the project definition
    """
    args = ['--quiet']
    args.append('update')
    args.append(to)
    if develop:
        args.append('--develop')
    executed, filename = _version.main(leave=False, args=args)

    if not executed:
        if filename == "":
            error("No project definition found.")
        else:
            error(f"Unable to update version {to} in {filename}")

    return executed, filename


def upload_assets(
    username: str,
    token: str,
    pathnames: List[str],
    github_json: Dict,
    path: Path,
    requests_module: requests,
) -> bool:
    """Function to upload assets

    Arguments:
        username: The GitHub username to use for the upload
        token: That username's GitHub token
        pathnames: List of paths to asset files
        github_json: The github dictionary, containing relevant information
            for the uplaod
        path: the python pathlib.Path module
        requests_module: the python request module

    Returns:
        True on success, false else
    """
    info(f"Uploading assets: {pathnames}")

    asset_url = github_json['upload_url'].replace('{?name,label}', '')
    paths = [path(f'{p}.asc') for p in pathnames]

    headers = {
        'Accept': 'application/vnd.github.v3+json',
        'content-type': 'application/octet-stream',
    }
    auth = (username, token)

    for path in paths:
        to_upload = path.read_bytes()
        resp = requests_module.post(
            f"{asset_url}?name={path.name}",
            headers=headers,
            auth=auth,
            data=to_upload,
        )

        if resp.status_code != 201:
            error(
                f"Wrong response status {resp.status_code}"
                f" while uploading {path.name}"
            )
            out(json.dumps(resp.text, indent=4, sort_keys=True))
            return False
        else:
            ok(f"uploaded: {path.name}")

    return True



# Copyright (C) 2019-2021 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Script to update the year of last modification
in the license header of source code files.\n
Also it appends a header if it is missing in the file.
"""

import sys
import os
import re

from argparse import ArgumentParser, Namespace
from datetime import datetime
from glob import glob

from subprocess import CalledProcessError, run
from typing import Dict, Tuple, Union
from pathlib import Path

SUPPORTED_FILE_TYPES = [
    ".bash",
    ".c",
    ".h",
    ".cmake",
    ".js",
    ".nasl",
    ".po",
    ".py",
    ".sh",
    ".txt",
    ".xml",
    ".xsl",
]
SUPPORTED_LICENCES = [
    "AGPL-3.0-or-later",
    "GPL-2.0-only",
    "GPL-2.0-or-later",
    "GPL-3.0-or-later",
]


def _get_modified_year(f: Path) -> str:
    """In case of the changed arg, update year to last modified year"""
    try:
        cmd = ["git", "log", "-1", "--format=%ad", "--date=format:%Y", str(f)]
        proc = run(
            cmd,
            text=True,
            capture_output=True,
            check=True,
            universal_newlines=True,
        )
        return proc.stdout.rstrip()
    except CalledProcessError as e:
        raise e


def _find_copyright(
    line: str,
    regex: re.Pattern,
) -> Tuple[bool, Union[Dict[str, Union[str, None]], None]]:
    """Match the line for the regex"""
    copyright_match = re.search(regex, line)
    if copyright_match:
        return (
            True,
            {
                "creation_year": copyright_match.group(1),
                "modification_year": copyright_match.group(2),
                "company": copyright_match.group(3),
            },
        )
    return False, None


def _add_header(suffix: str, licence: str, company: str) -> Union[str, None]:
    """Tries to add the header to the file.
    Requirements:
      - file type must be supported
      - licence file must exist
    """
    if suffix in SUPPORTED_FILE_TYPES:
        root = Path(__file__).parent
        licence_file = root / "templates" / licence / f"template{suffix}"
        try:
            return licence_file.read_text().replace("Company", company)
        except FileNotFoundError as e:
            raise e
    else:
        raise ValueError


def _update_file(
    file: Path,
    regex: re.Pattern,
    args: Namespace,
) -> None:
    """Function to update the given file.
    Checks if header exists. If not it adds an
    header to that file, else it checks if year
    is up to date
    """

    if args.changed:
        try:
            args.year = _get_modified_year(file)
        except CalledProcessError:
            print(
                f"{file}: Could not get date of last modification"
                f" using git, using {str(args.year)} instead."
            )

    try:
        with file.open("r+") as fp:
            found = False
            i = 10  # assume that copyright is in the first 10 lines
            while not found and i > 0:
                line = fp.readline()
                if line == "":
                    i = 0
                    continue
                found, copyright_match = _find_copyright(line=line, regex=regex)
                i = i - 1
            if i == 0 and not found:
                try:
                    header = _add_header(
                        file.suffix, args.licence, args.company
                    )
                    if header:
                        fp.seek(0)  # back to beginning of file
                        rest_of_file = fp.read()
                        fp.seek(0)
                        fp.write(header)
                        fp.write(rest_of_file)
                        print(f"{file}: Added licence header.")
                        return 0
                except ValueError:
                    print(
                        f"{file}: No licence header for the"
                        f" format {file.suffix} found.",
                    )
                except FileNotFoundError:
                    print(
                        f"{file}: Licence file for {args.licence} "
                        "is not existing."
                    )
                return 1
            # replace header and write it to file
            if (
                not copyright_match["modification_year"]
                and copyright_match["creation_year"] < args.year
                or copyright_match["modification_year"]
                and copyright_match["modification_year"] < args.year
            ):
                copyright_term = (
                    f'Copyright (C) {copyright_match["creation_year"]}'
                    f'-{args.year} {copyright_match["company"]}'
                )
                new_line = re.sub(regex, copyright_term, line)
                fp_write = fp.tell() - len(line)  # save position to insert
                rest_of_file = fp.read()
                fp.seek(fp_write)
                fp.write(new_line)
                fp.write(rest_of_file)
                print(
                    f"{file}: Changed Licence Header Copyright Year "
                    f'{copyright_match["modification_year"]} -> '
                    f"{args.year}"
                )

                return 0
            else:
                print(f"{file}: Licence Header is ok.")
                return 0
    except FileNotFoundError as e:
        print(f"{file}: File is not existing.")
        raise e
    except UnicodeDecodeError as e:
        print(f"{file}: Ignoring binary file.")
        raise e


def _parse_args(args=None):
    """Parsing the args"""

    parser = ArgumentParser(
        description="Update copyright in source file headers.",
        prog="pontos-copyright",
    )

    date_group = parser.add_mutually_exclusive_group()
    date_group.add_argument(
        "-c",
        "--changed",
        action="store_true",
        default=False,
        help=(
            "Update modified year using git log modified year. \n"
            "This will probably not changed all files to current year!",
        ),
    )
    date_group.add_argument(
        "-y",
        "--year",
        default=str(datetime.now().year),
        help=(
            "If year is set, modified year will be \n"
            "set to the specified year."
        ),
    )

    parser.add_argument(
        "-l",
        "--licence",
        choices=SUPPORTED_LICENCES,
        default="GPL-3.0-or-later",
        help="Add header f files",
    )

    parser.add_argument(
        "--company",
        default="Greenbone Networks GmbH",
        help=(
            "If a header will be added to file, \n"
            "it will be licenced by company."
        ),
    )

    files_group = parser.add_mutually_exclusive_group(required=True)
    files_group.add_argument(
        "-f", "--files", nargs="+", help="Files to update."
    )
    files_group.add_argument(
        "-d",
        "--directory",
        help="Directory to find files to update recursively.",
    )
    return parser.parse_args(args)


def main() -> None:
    args = _parse_args()

    if args.directory:
        # get files to update
        files = [
            Path(file)
            for file in glob(args.directory + "/**/*", recursive=True)
            if os.path.isfile(file)
        ]
    elif args.files:
        files = [Path(name) for name in args.files]

    else:
        # should never happen
        print("Specify files to update!")
        sys.exit(1)

    regex = re.compile(
        "[Cc]opyright.*?(19[0-9]{2}|20[0-9]{2}) "
        f"?-? ?(19[0-9]{{2}}|20[0-9]{{2}})? ({args.company})"
    )

    for file in files:
        try:
            _update_file(file=file, regex=regex, args=args)
        except (FileNotFoundError, UnicodeDecodeError, ValueError):
            continue

    return 0


if __name__ == "__main__":
    main()
